local ADDON_NAME = 'Personal Loot Trader'
local addonName, addon = ...
local PLTrader = addon
local PLT = PLTrader

PLT.VERSION = GetAddOnMetadata(addonName, "Version")
PLT.AUTHOR = GetAddOnMetadata(addonName, "Author")
PLT.ICONPATH = "Interface\\AddOns\\PersonalLootTrader\\media\\icons\\icons.tga"
local DF = _G ["DetailsFramework"]
local PersonalLootTrader = DF:CreateAddOn ("PersonalLootTrader", "PersonalLootTraderVariables", default_config)

if (not DF) then
	print ("|cFFFFAA00Personal Loot Trader: framework not found, if you just installed or updated the addon, please restart your client.|r")
	return
end
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub ("AceLocale-3.0"):GetLocale ("PersonalLootTrader", true)
if (not L) then
	print ("|cFFFFAA00Personal Loot Trader|r: Reopen your client to finish updating the addon.|r")
	return
end

local	SelfLootPattern = _G.LOOT_ITEM_SELF
		SelfLootPattern = SelfLootPattern:gsub('%%s', '(.+)')
		
local	SelfItemPattern = _G.LOOT_ITEM_PUSHED_SELF 
		SelfItemPattern = SelfItemPattern:gsub('%%s', '(.+)')

local	ItemLevelPattern = _G.ITEM_LEVEL
		ItemLevelPattern = ItemLevelPattern:gsub('%%d', '(%%d+)')  -- 'Item Level (%d+)'

local	RollPattern = _G.RANDOM_ROLL_RESULT --fix roll pattern to be auto
	    RollPattern = RollPattern:gsub('%%s', '(.+)')
	    RollPattern = RollPattern:gsub('%%d %(%%d%-%%d%)', '(%%d+) %%((%%d+)%%-(%%d+)%%)')
	    RollPattern = '^' .. RollPattern .. '$'

local MAIN_FRAME_WIDTH = 240
local MAIN_FRAME_HEIGHT = 180
local PATTERN_TRADE_TIME_REMAINING = BIND_TRADE_TIME_REMAINING:gsub("%%s", ".+")

local GetContainerNumSlots = GetContainerNumSlots --C_Container.GetContainerNumSlots
local GetContainerItemLink = GetContainerItemLink --C_Container.GetContainerItemLink

local version = VERSION
local debugging = false

local itemStorage = {}
local d = "~" --Delimiter
local tooltipName = "PLTTooltipScanner"
local PLTTooltipScanner = CreateFrame("GameTooltip", tooltipName, nil, "GameTooltipTemplate")
local rollStorage, timerLength

local inUseNumbers = {}
local PLT_FONT = "Interface\\AddOns\\PersonalLootTrader\\media\\fonts\\DORISBR.TTF"
local START_NUM = 200
local MSNum = START_NUM
local OSNum

C_ChatInfo.RegisterAddonMessagePrefix("PLTrader2")

--TODO

local EF = CreateFrame('Frame') -- event handler frame
EF:RegisterEvent('ADDON_LOADED')
EF:RegisterEvent('PLAYER_ENTERING_WORLD')

EF:SetScript('OnEvent', function(self, event, ...) self[event](self, ...) end)


function PLT:Print(text)
	print("|cFFFFAA00<PTL> |r"..text)
end

function PLT:DebugPrint(...)
    if not PLTrader.db.profile.general.debugging then return end
    print("PLTDB:", ... )
end

local function dbpr(...)
    PLT:DebugPrint(...)
end

function PLT:Round(number, decimals)
    return (("%%.%df"):format(decimals)):format(number)
end

function PLT:ClearUseNumber(num)
	num = tonumber(num)
	PLT:DebugPrint(num.. " is now cleared")
	inUseNumbers[num] = nil
	PLT:UpdateExampleText()
end

function PLT:AddUseNumber(num)
	num = tonumber(num)
	PLT:DebugPrint(num.. " is now in use.")
	inUseNumbers[num] = true
	PLT:UpdateExampleText()
end

function PLT:IsNumInUse(num)
	--if inUseNumbers[num] then return true end
    return inUseNumbers[num]
end

function PLT:Initialize(event, addon)
   	self.defaults = {
		profile = {
			general = {
				scale = 1,
				version = VERSION,
				debugging = false,
				tooltips = true,
			},
			trader = {
				usetimer = false,
				length = 60,
				autoshow = true,
				xoff = (GetScreenWidth()/2)-(MAIN_FRAME_WIDTH/2),
				yoff = (GetScreenHeight()/2)+(MAIN_FRAME_HEIGHT/2),
				announcewinner = true,
				whisperwinner = false,
				MSOSMessage = L["S_ROLL_MSG"],
				msabbrev = L["S_MainSpecAbreviation"],
				osabbrev = L["S_OffSpecAbreviation"],
				MSMessage = L["S_MS_ROLL_MSG"],
				allowOS = true,
				autoAssign = true,
				
			},
			roller = {
				xoff = (GetScreenWidth()/4),
				yoff = (GetScreenHeight()-GetScreenHeight()/4),
				receive = true,
				autopass = true,
				length = 60,
				corpse_arrow = true,
				show = true,
				searchchat = true,
				searchchattime = 30,
				currentspec = false,
				test = true,
			},
            rclc = {
                replace = false,
                stopplt = true,
            }
		},
	}
	
	PLT.name = GetUnitName("player") .. "-" .. GetRealmName();
	PLT.specIDs = {}
	local _, _, classIndex = UnitClass("player");
	for i=1, GetNumSpecializationsForClassID(classIndex) do
		local specID = GetSpecializationInfoForClassID(classIndex, i)
		PLT.specIDs[tostring(specID)] = true
	end
	
	self.defaults.profile.roller[PLT.name] = {}
	
	for ID, _ in pairs(PLT.specIDs) do
		self.defaults.profile.roller[PLT.name][ID] = true
	end
	
	self.db = LibStub("AceDB-3.0"):New("PLTraderDB", self.defaults, "Default")
end

local function isDebugging()
	if PLTrader.db.profile.general.debugging then
		return true
	end
end
 
local function shuffle(tbl)
  size = #tbl
  for i = size, 1, -1 do
    local rand = math.random(size)
    tbl[i], tbl[rand] = tbl[rand], tbl[i]
  end
  return tbl
end

function PLT:NS(name)
	name = gsub(name, "%-[^|]+", "")
	return name
end

function PLT:CreateEmptyTooltip(rows)
    local tip = CreateFrame('GameTooltip', "PLTScanningTooltip", nil, "GameTooltipTemplate")
	local leftside = {}
	local rightside = {}
	local L, R
    for i = 1, rows do
        L, R = tip:CreateFontString(), tip:CreateFontString()
        L:SetFontObject(GameFontNormal)
        R:SetFontObject(GameFontNormal)
        tip:AddFontStrings(L, R)
        leftside[i] = L
		rightside[i] = R
    end
    tip.leftside = leftside
	tip.rightside = rightside
    return tip
end

function PLT:GetILVLFromTooltip(tooltip)
	local ITEM_LEVEL_PATTERN = _G.ITEM_LEVEL:gsub('%%d', '(%%d+)')  				-- Item Level (%d+)
	local ilvl = nil
	local text = _G["PLTScanningTooltipTextLeft2"]:GetText()--tooltip.leftside[2]:GetText()
    if text ~= nil then
		ilvl = text:match(ITEM_LEVEL_PATTERN)
	end
	if ilvl == nil then  -- ilvl can be in the 2nd or 3rd line dependng on the tooltip; if we didn't find it in 2nd, try 3rd
		text =  _G["PLTScanningTooltipTextLeft3"]:GetText()--tooltip.leftside[3]:GetText()
        if text ~= nil then
			ilvl = text:match(ITEM_LEVEL_PATTERN)
		end
	end
	return ilvl
end

function PLT:DebugRollAdd(maxRoll)
	-- if not PLTrader.testName then
		-- PLTrader.testName = 0
	-- else
		-- PLTrader.testName = PLTrader.testName + 1
	-- end
	PLTrader.testName = math.random (1, 8)
	local rollNum = math.random (maxRoll-2, maxRoll)
	--local rollTbl = {"name"..PLTrader.testNum,rollNum,1,100}
	EF:CHAT_MSG_SYSTEM("PlayerTest"..PLTrader.testName.." rolls "..rollNum.." (1-"..maxRoll..")")
	
	-- EF:CHAT_MSG_SYSTEM("Player1 rolls 199 (1-200)")
	-- EF:CHAT_MSG_SYSTEM("Player2 rolls 199 (1-200)")
	-- EF:CHAT_MSG_SYSTEM("Player1 rolls 198 (1-200)")
	-- EF:CHAT_MSG_SYSTEM("Player2 rolls 167 (1-200)")
	-- EF:CHAT_MSG_SYSTEM("Player3 rolls 199 (1-200)")
end

local function isBagSlotPersonalLootTradeable(bag, slot)
	PLTTooltipScanner:SetOwner(UIParent, "ANCHOR_NONE")
	PLTTooltipScanner:SetBagItem(bag, slot)
	local tipRegion, tipText
	for k = 1, PLTTooltipScanner:NumLines() do
		tipRegion = _G["PLTTooltipScannerTextLeft" .. k]
		tipText = tipRegion:GetText()
		
		if tipText:match(PATTERN_TRADE_TIME_REMAINING) then
			PLT:DebugPrint("It's personal loot and tradeable")
			return true
		end
	end

end

function PLT:IsTradeable(itemLink)     
	-- verify item exists in bags first
	local bag, slot
	for i = 0, NUM_BAG_SLOTS do -- NUM_BAG_SLOTS is an API constant
		for j = 1, GetContainerNumSlots(i) do
			if GetContainerItemLink(i, j) == itemLink then
				
				PLT:DebugPrint("Item is in your bag")
				
				bag, slot = i, j
				
				local canTrade = isBagSlotPersonalLootTradeable(bag, slot)

				if canTrade then
					return true, bag, slot
				end
			end
		end
	end
	
	local _, _, _, _, _, _, _, _, _, _, _, _, _, bindType, _, _, _ = GetItemInfo(itemLink)
	if bindType and bag then
		if bindType == 2 or bindType == 0 then
			return true, bag, slot, true
		else
			return false, bag, slot
		end
	end
end

local function PLTIsOpen()
	if PLT.MF and PLT.MF:IsVisible() then
		return true
	end
end

local function GetRollSplit(fullstr)
	local num, _ = strsplit(";", fullstr, 2)
	return tonumber(num)
end

local function GetNameSplit(fullstr)
	if fullstr then
		local _, name = strsplit(";", fullstr, 2)
		return name
	else
		return nil
	end
end

local function ShowWinnerButton(RerollShow) --forces tooltip update when hovering over
	PLT.MF.PickWinnerButton:Show()
	
	if RerollShow then
		PLT.MF.Reroll:Show()
	end	
end

function PLT:RegisterDefaultEvents()
	EF:UnregisterAllEvents()
	if IsInGroup() then
		PLT:DebugPrint("Registering Loot")
		EF:RegisterEvent('CHAT_MSG_LOOT')
		EF:RegisterEvent('ITEM_LOCKED')
	end
	EF:RegisterEvent('GROUP_JOINED')
	EF:RegisterEvent('GROUP_LEFT')
	EF:RegisterEvent('PLAYER_LOGOUT')
	EF:RegisterEvent('UI_SCALE_CHANGED')
end

function EF:ADDON_LOADED(addon)
	if addon == "PersonalLootTrader" then
		EF:UnregisterEvent("ADDON_LOADED")
		PLT:Initialize()
		
		PLTraderDB.version = PLT.VERSION
		
		PLT:UpdateChatEvents()
		PLT:ChangeReceive()
		
		--Fonts confuse me so uhh hope this isn't terrible
		PLT.font = CreateFont("PLTButtonFont")
		PLT.font:CopyFontObject("GameFontNormal");
		PLT.font:SetTextColor(1, 1, 1, 1.0);
		PLT.font:SetJustifyH("CENTER")
		PLT.font:SetFont(PLT_FONT, 12, "")
		-- PLT:CreateRollerFrames()
	end
end

function EF:ITEM_LOCKED(...)

	if PLT.MF and PLT.MF:IsVisible() then return end 
	local bag, slot = ...

	if (bag and bag == -1) or not slot then return end --exit if you are dragging an equipped item or a bag in the bank
	
	local itemLink = GetContainerItemLink(bag, slot)

    if not itemLink then return end

	local _, _, itemRarity = GetItemInfo(itemLink)
	
	local canTrade = isBagSlotPersonalLootTradeable(bag, slot)
	
	if canTrade then
		-- itemStorage.curItemLink = itemLink
		-- PLT:OpenMainFrame(true)
		itemStorage.draggingLink = itemLink
		PLT:OpenDragFrame()
		EF:RegisterEvent("GLOBAL_MOUSE_UP")
		EF:RegisterEvent("ITEM_UNLOCKED")
		PLT:DebugPrint("You can trade the item you're dragging.")
	end
end

function PLT:DCUpdate(entered)
	if entered and itemStorage.draggingLink then
		local _, _, itemRarity = GetItemInfo(itemStorage.draggingLink)
		if itemRarity then
			r, g, b, hex = GetItemQualityColor(itemRarity)		
			PLT.MF.DC.LootBorder:SetBackdropColor (r, g, b, .2)
			PLT.MF.DC.LootBorder:SetBackdropBorderColor (r, g, b, 1)
		end
	else
		PLT.MF.DC.LootBorder:SetBackdropColor (0, 0, 0, .8)
		PLT.MF.DC.LootBorder:SetBackdropBorderColor (1, 1, 1, .5)
	end
end

function EF:UI_SCALE_CHANGED()
	--print(uiScale)
	-- PLTrader.db.profile.trader.xoff = PLTrader.db.profile.trader.xoff/uiScale
	-- PLTrader.db.profile.trader.yoff = PLTrader.db.profile.trader.xoff/uiScale
	
	-- PLT:UpdateFramePositions()
end

function EF:ITEM_UNLOCKED()
	EF:UnregisterEvent("ITEM_UNLOCKED")
	PLT:HideMainFrame()
end

function EF:GLOBAL_MOUSE_UP()
	EF:UnregisterEvent("GLOBAL_MOUSE_UP")
	EF:UnregisterEvent("ITEM_UNLOCKED")
	if MouseIsOver(PLT.MF.DC.LootBorder) then
		itemStorage.curItemLink = itemStorage.draggingLink
		PLT:OpenMainFrame(true)
		ClearCursor()
		CloseAllBags()
	else
		PLT.MF:Hide()
	end
end

local function addLootRCLCWinner(itemLink, winner)
    if itemLink and winner then
        PLT:CheckItem(itemLink, true)
        PLT.MF.fromRCLC = true
        PLT:CustomWinner(PLT:NS(winner))
        RCLootCouncil.TradeUI:Hide()
    end
end

local function RCLCTradeReceived(...)

    if not PLTrader.db.profile.rclc.replace then return end

    if PLT.MF and PLT.MF:IsVisible() then return end

    local _, session, winner, trader = ...

    if PLT:NS(trader) == PLT:NS(UnitName("player")) then
        dbpr("PLAYER IS TRADER")
        local lootSession = RCLootCouncil:GetLootTable()[session]
        --addon.Log:d("OnAwardReceived", lootSession, session, winner, trader)
        if lootSession then
            dbpr("IS LOOT SESSION")
            --Item = getItemFromLootSession(lootSession.link, session, winner)
            --Item.type = "to_trade"
            --Item.args.recipient = winner
            --Item.args.session = session
            dbpr(lootSession.link)
            addLootRCLCWinner(lootSession.link, winner)
        else
            dbpr("NOT LOOT SESSION")
            -- If the session has ended and we receive another award, it's got to be a reaward. Fetch the item based on the session:
            local Items = RCLootCouncil.TradeUI:GetStoredItemBySession(session)
            if Items and #Items > 0 then

               Item = Items[1]
            --   Item.args.recipient = winner
            --   addon.Log:d("Changed winner for", session, Item.link, "to", winner)

                dbpr(Item.link)
                addLootRCLCWinner(Item.link, winner)
            end
        
        end

    end
end

--local function RCLCClickedKeep(...)
--    local _, link, owner, reason = ...

--    dbpr("RC Clicked Keep",link, owner, reason)
--    if not PLTrader.db.profile.rclc.stopplt then return end

--    if not (PLT.MF and PLT.MF:IsVisible()) then return end


--    if PLT:NS(owner) == select(1, UnitName("player")) and reason == "r_t" then
--        PLT:HideMainFrame()
--        PLT:Print(L["S_RCLC_HIDE"])
--    end
--end

--local function RCLCClickedTrade(...)
--    local _, sender, item = ...

--    dbpr("RC Clicked Trade", sender, item)

--    if not PLTrader.db.profile.rclc.stopplt then return end

--    if not (PLT.MF and PLT.MF:IsVisible()) then return end

--    if PLT:NS(sender) == select(1, UnitName("player")) then
--        PLT:HideMainFrame()
--        PLT:Print(L["S_RCLC_HIDE"])
--    end
--end

--local function LibDialogSpawn(...)
--    local dialogName = select(2, ...)
--    if dialogName == "RCLOOTCOUNCIL_KEEP_ITEM" then
--        if not PLTrader.db.profile.rclc.stopplt then return end

--        if not (PLT.MF and PLT.MF:IsVisible()) then return end

--        PLT:HideMainFrame()
--        PLT:Print(L["S_RCLC_HIDE"])

--    end
--end

function EF:PLAYER_ENTERING_WORLD()
	PLT:RegisterDefaultEvents()
    if RCLootCouncil then
        local LibDialog = LibStub("LibDialog-1.0")
        hooksecurefunc(RCLootCouncil.TradeUI, "OnAwardReceived", RCLCTradeReceived)

        --RCLootCouncil doesn't have a good way to check if a session is running..

        --hooksecurefunc(RCLootCouncilML, "HandleNonTradeable", RCLCClickedKeep)
        --hooksecurefunc(RCLootCouncilML, "HandleReceivedTradeable", RCLCClickedTrade)
        --hooksecurefunc(LibDialog, "Spawn", LibDialogSpawn)

    end

	EF:UnregisterEvent('PLAYER_ENTERING_WORLD')
end

function EF:GROUP_JOINED()
	PLT:DebugPrint("Registering Loot")
	EF:RegisterEvent('CHAT_MSG_LOOT')
	EF:RegisterEvent('ITEM_LOCKED')
end

function EF:GROUP_LEFT()
	if IsInGroup() then return end -- You can leave an instance group and still be in a group. Does this fix it?
	PLT:DebugPrint("Unregistering Loot")
	EF:UnregisterEvent('CHAT_MSG_LOOT')
	EF:UnregisterEvent('ITEM_LOCKED')
end

local function pairsByKeys (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a,  function(a,b) return b<a end)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

local function updateTies(player, rollType)

	local tieDB = rollStorage.playerRollDB.ties[rollType]
	
	local curPlayerRollDB = rollStorage.playerRollDB[rollType][player]
	local i = 0

	local tie = true
	
	while tie do
		i = i + 1
		
		local prevPlayer = rollStorage.sortedRolls[#rollStorage.sortedRolls-i].player
		
		if not prevPlayer then return end
		
		local prevPlayerRollDB = rollStorage.playerRollDB[rollType][prevPlayer]
		
		for i, roll in ipairs(curPlayerRollDB) do
			if prevPlayerRollDB[i] ~= roll then
				tie = false
			end
		end
		
		if tie then
			if not tieDB[player] then tieDB[player] = {} end
			tieDB[player][prevPlayer] = true
			if not tieDB[prevPlayer] then tieDB[prevPlayer] = {} end
			tieDB[prevPlayer][player] = true
		end
	end
	
end

local function sortRolls(t, rollType)
	if t.rolls then
		for roll, t in pairsByKeys(t.rolls) do
			sortRolls(t, rollType)
		end
	end

	if not t.players then return end
	for player, _ in pairs(t.players) do
		if not rollStorage.ignoreNames[player] then
			rollStorage.ignoreNames[player] = true --Already found their highest roll. Names still exist in previous rolls so ignore them.
			local sortData = {player = player, rollType = rollType}
			table.insert(rollStorage.sortedRolls, sortData)
			
			updateTies(player, rollType)
			
		end
	end

end


function PLT:HidePlayerButtons()
	for _, button in pairs(PLT.MF.PlayerButtons) do
		button:Hide()
	end
end

function PLT:UncheckPlayerButtons()
	for _, button in pairs(PLT.MF.PlayerButtons) do
		button:SetChecked(false)
	end
end

function PLT:InitiateTrade()
	local inRange = CheckInteractDistance(rollStorage.winner, 2);
	EF:RegisterEvent("TRADE_CLOSED")
	--TargetUnit(PLT:NS(Winner));
	if inRange then
		InitiateTrade(rollStorage.winner);
	else
		inRange = CheckInteractDistance(rollStorage.winner, 4);
		if inRange then
			FollowUnit(rollStorage.winner)
		end
	
	end
end

function PLT:UpdateTradeButton(player, button)
	
	player = PLT:NS(player)
	
	local inRange = CheckInteractDistance(player, 2);
	
	if inRange then
		button:SetText(L["S_INITIATE_TRADE"])
		PLT:ButtonState(button, true)
	else
		inRange = CheckInteractDistance(player, 4);
		button:SetText(L["S_FOLLOW"])
		if inRange then
			PLT:ButtonState(button, true)	
		else
			PLT:ButtonState(button, false)
		end

	end

end

function PLT:UpdateTraderTradeButton()
	if not PLT.MF:IsVisible() then return end
	
	local inRange = CheckInteractDistance(rollStorage.winner, 2);
	
	PLT:UpdateTradeButton(rollStorage.winner, PLT.MF.Trade)
	if rollStorage.awaitingTrade then
		C_Timer.After(0.3, function() PLT:UpdateTraderTradeButton() end)
	else
		--font:SetTextColor(1, 1, 1, 1);
	end
	
end

function PLT:WhisperPlayer(player, text)
	if rollStorage.playerRealm[player] then
		player = player .. "-" .. rollStorage.playerRealm[player]
	end
	SendChatMessage(text, "WHISPER", nil, player)
end

local function getRerollNames(tieDB) --Pretty bad but don't feel like making better
	local maxRollSize = 0
	
	for player, _ in pairs(tieDB) do
		local rollDB = rollStorage.playerRollDB[rollStorage.selectedRollType][player]
		if #rollDB > maxRollSize then
			maxRollSize = #rollDB
		end
	end
	local rollDB = rollStorage.playerRollDB[rollStorage.selectedRollType][rollStorage.selectedPlayer]
	if #rollDB > maxRollSize then
		maxRollSize = #rollDB
	end
	
	local rerollNames = {}
	
	for player, _ in pairs(tieDB) do
		local rollDB = rollStorage.playerRollDB[rollStorage.selectedRollType][player]
		if #rollDB < maxRollSize then
			table.insert(rerollNames, player)
		end
	end

	local rollDB = rollStorage.playerRollDB[rollStorage.selectedRollType][rollStorage.selectedPlayer]
	if #rollDB < maxRollSize then
		table.insert(rerollNames, rollStorage.selectedPlayer)
	end
	
	--If everyone is tied then everyone needs to reroll 
	if #rerollNames > 0 then
		return rerollNames
	else
		for player, _ in pairs(tieDB) do
			table.insert(rerollNames, player)
		end
		table.insert(rerollNames, rollStorage.selectedPlayer)
		return rerollNames
	end
end

local function getTieDB() --Probably should make a table/variable that holds the current tieDB
	local tieDB = rollStorage.playerRollDB.ties[rollStorage.selectedRollType]
	if tieDB then 
		tieDB = tieDB[rollStorage.selectedPlayer] 
		return tieDB
	end
end


function PLT:PickWinnerPress(customWinner)
	if PLT.MF.hideTimer then return end
	local tieDB = getTieDB()
	if not tieDB then --not a tie
		if rollStorage.winner == rollStorage.selectedPlayer then return end
		PLT.MF.manualSelect = true
		rollStorage.winner = rollStorage.selectedPlayer
		PLT:UpdateNameWinnerSpec()
		
		local db = self.db.profile.trader
		
		if not customWinner then
			if db.announcewinner then
				SendChatMessage(L["S_ADDON_ABREV"].." "..format(L["S_WINNER_ANNOUNCE"], rollStorage.winner, itemStorage.curItemLink), PLTrader.channel)
			end
			
			if db.whisperwinner then
				PLT:WhisperPlayer(rollStorage.winner, format(L["S_OPEN_TRADE"], itemStorage.curItemLink))
			end
			
			local message = "End"..d..rollStorage.winner..d..MSNum..d..itemStorage.curItemLink
			--local message = "End~"..rollStorage.winner.."~"..MSNum
			C_ChatInfo.SendAddonMessage("PLTrader2", message, PLTrader.channel);
		end
	
		PLT:UpdatePickWinnerButtonText(false)
		PLT.MF.Trade:Show()

		EF:RegisterEvent("TRADE_SHOW")
		rollStorage.awaitingTrade = true
		PLT:UpdateTraderTradeButton()
	else --it's a tie
		local rerollNames = getRerollNames(tieDB)
		for _, name in pairs(rerollNames) do
			PLT:DebugPrint(name)

			local rollNum
			
			if rollStorage.selectedRollType == "MS" then
				rollNum = MSNum
			else
				rollNum = OSNum
			end
			
			local message = format (L["S_TIE_WHISPER"], rollNum)
			message = message:gsub(" 100", "")
			PLT:WhisperPlayer(name, message)
		end
	end

end

local function inspectGroupMember(name)
	if name then --and name ~= UnitName('player') then 
		if CanInspect(name) then
			if UnitIsVisible(name) then
				NotifyInspect(name)
				return true
			else
				PLT:DebugPrint(name .. ' out of range for inspect')
			end
		else
			PLT:DebugPrint('Unable to inspect ' .. name)
		end
	end
	return false
end

function PLT:PlayerSelected(tie)
	
	PLT:UpdateNameWinnerSpec(tie)
	
	if not rollStorage.playerSpec[rollStorage.selectedPlayer] then

		--If we don't know the spec then initiate an inspect
		local tryInspect = inspectGroupMember(rollStorage.selectedPlayer)
		if tryInspect then
			PLT:DebugPrint("Inspect Registered")
			EF:RegisterEvent("INSPECT_READY")
		end
	end
	
	PLT:UpdatePickWinnerButtonText(tie)
end

function PLT:PlayerButtonPress(button)
	if not button.playerName then return end
	
	PLT.MF.manualSelect = true
	
	PLT:UncheckPlayerButtons()
	rollStorage.selectedPlayer = button.playerName
	rollStorage.selectedRollType = button.rollType
	button:SetChecked(true)
	
	local tieDB = getTieDB()
	
	PLT:PlayerSelected(tieDB)
	
	--If there is a tie we need to check all of the tied players
	if not tieDB then return end
	
	for _, button in pairs(PLT.MF.PlayerButtons) do
		if tieDB[button.playerName] and button.rollType == rollStorage.selectedRollType then button:SetChecked(true) end 
	end
	
	
end

local function getPlayerButtonText(button)

	local playerRolls
	if button.rollType == "MS" then
		playerRolls = rollStorage.playerRollDB.MS[button.playerName]
	else
		playerRolls = rollStorage.playerRollDB.OS[button.playerName]
	end
	
	local rollText = ""
	for i, roll in ipairs(playerRolls) do
		if rollText ~= "" then
			rollText = rollText .. "-" .. roll
		else
			rollText = roll
		end
	end
	
	local text = "    [".. rollText.."] "..PLT:ColorName(PLT:NS(button.playerName))
	
	return text--PLT.MF.PlayerButtons[i]:SetText(text)
end

function PLT:GetRaidMemberClasses()
	for i=0, GetNumGroupMembers() do
		local name, _, _, _, class = GetRaidRosterInfo(i);
		if name then
			rollStorage.playerClass[PLT:NS(name)] = class
		end
	end

end 

local function colorNameFromClass(name, class)
	local color = select(4, GetClassColor(strupper(class)))
	local coloredName = "|c"..color..name.."|r"
	return coloredName
end

function PLT:ColorName(name)
	if not name then return "<Error>" end
	if rollStorage.playerClass[name] then
		local class = string.gsub(rollStorage.playerClass[name], " ", "")
		local coloredName = colorNameFromClass(name, class)
		return coloredName
	else
		return "|cff808080".. name .. "|r" --Colors people with no class gray. Can't get class info for some reason?
	end
end


local PLAYER_BUTTON_HEIGHT = 15
local PLAYER_BUTTON_PADDING = 1
local NUM_BUTTONS = 5

local function updateScrollFrame(self)
	if not self then return end
	local list = rollStorage.sortedRolls
	local numItems = #list
	FauxScrollFrame_Update(self, numItems, NUM_BUTTONS, PLAYER_BUTTON_HEIGHT)
	local offset = FauxScrollFrame_GetOffset(self)
	
	local tieDB = getTieDB()
	
	for line = 1, NUM_BUTTONS do
		local linePlusOffset = line + offset
		local button = PLT.MF.PlayerButtons[line]
		if linePlusOffset > numItems then
			button:Hide()
		else
			button:SetChecked(false)
			if not rollStorage.sortedRolls[linePlusOffset].player then
				button.playerName = nil
				button:SetSize(.1, PLAYER_BUTTON_HEIGHT)
				button:SetText(rollStorage.sortedRolls[linePlusOffset]..":")
			else
				button:SetSize(PLAYER_BUTTON_HEIGHT, PLAYER_BUTTON_HEIGHT)
				
				button.playerName = rollStorage.sortedRolls[linePlusOffset].player
				button.rollType = rollStorage.sortedRolls[linePlusOffset].rollType
				
				button:SetText(getPlayerButtonText(button))
				
				--Set checked if it's the selected player or if the person is tied with selected player
				if button.rollType == rollStorage.selectedRollType then
					if (button.playerName == rollStorage.selectedPlayer) or (tieDB and tieDB[button.playerName]) then button:SetChecked(true) end
				end
				
				--button:SetText(list[linePlusOffset])
			end
			button:Show()
		end
	end
	
end	


local function eraseTies(rollType)
	rollStorage.playerRollDB.ties[rollType] = {}
end

local function sortRollFromTable(rollType)
	local t = rollStorage[rollType.."Rolls"]
	eraseTies(rollType)
	if t.rolls then
		table.insert(rollStorage.sortedRolls, PLTrader.db.profile.trader[string.lower(rollType).."abbrev"] .. " ["..(rollStorage.playerRollDB[rollType].rollCount).."]")
		sortRolls(t, rollType)
	end
	
	rollStorage.ignoreNames = {}
end

local function createScrollFrame()
	PLT.MF.ScrollFrame = CreateFrame("ScrollFrame", "PLTScrollFrame", PLT.MF.RC, "FauxScrollFrameTemplate")
	PLT.MF.ScrollFrame:SetScript("OnVerticalScroll", function(self, offset)
		FauxScrollFrame_OnVerticalScroll(self, offset, PLAYER_BUTTON_HEIGHT, updateScrollFrame)
	end)
	PLT.MF.ScrollFrame:SetPoint("TOPLEFT", PLT.MF.TitleBar, "BOTTOMLEFT", 5, -5)
	PLT.MF.ScrollFrame:SetWidth(200)
	PLT.MF.ScrollFrame:SetHeight(NUM_BUTTONS * PLAYER_BUTTON_HEIGHT)
	DF:ReskinSlider (PLT.MF.ScrollFrame)

	for i = 1, NUM_BUTTONS do
		local button = CreateFrame("CheckButton", nil, PLT.MF.RC, "UIRadioButtonTemplate")
		button:SetNormalFontObject(PLT.MF.playerFont)
		button:SetHitRectInsets(0,-200,0,0)

		button:SetChecked(false)
		button:SetScript ("OnClick", function() PLT:PlayerButtonPress(button) end)
		
		if i == 1 then
			button:SetPoint("TOPLEFT", PLT.MF.ScrollFrame, "TOPLEFT", 30)
		else
			button:SetPoint("TOPLEFT", PLT.MF.PlayerButtons[i - 1], "BOTTOMLEFT")
		end
		button:SetSize(PLAYER_BUTTON_HEIGHT, PLAYER_BUTTON_HEIGHT)
		PLT.MF.PlayerButtons[i] = button
	end
end

function PLT:CreateSortedNameList()

	rollStorage.sortedRolls = {}
	rollStorage.ignoreNames = {}
	
	sortRollFromTable("MS")
	sortRollFromTable("OS")
	
	if not PLT.MF.manualSelect then --If you haven't selected a player manually, select the top
		rollStorage.selectedPlayer = rollStorage.sortedRolls[2].player
		rollStorage.selectedRollType = rollStorage.sortedRolls[2].rollType
		PLT:DebugPrint("Auto Picking " .. rollStorage.selectedPlayer)
	end

	
	PLT:PlayerSelected(getTieDB())
	
	if not PLT.MF.ScrollFrame then
		createScrollFrame()
	end
	
	updateScrollFrame(PLT.MF.ScrollFrame)

end

function PLT:CreatePlayerButtons()
	--GPORM:HideBossButtons() --Hide all the buttons so if you import a smaller boss list extra buttons aren't there
	for i, rollInfo in ipairs(rollStorage.sortedRolls) do
		if not PLT.MF.PlayerButtons[i] then
			 PLT:CreatePlayerButton(i)
		end
		PLT:UpdatePlayerButton(i)
	end
end

function PLT:CreatePlayerButton(i)

	local button = CreateFrame("button", "PLTPlayerButton"..i, PLT.MF.PBC)
	if i == 1 then
		button:SetPoint("TOPLEFT", PLT.MF.PBC, "TOPLEFT", 20, -3)
	else
		button:SetPoint("TOPLEFT", PLT.MF.PlayerButtons[i-1], "BOTTOMLEFT", 0, -1 * PLAYER_BUTTON_PADDING)
	end	
	
	button:SetScript("OnClick", function() PLT:PlayerButtonPress(PLT.MF.PlayerButtons[i]) end)
	
	button:SetWidth(100)
	button:SetHeight(PLAYER_BUTTON_HEIGHT)

	button:SetNormalFontObject(PLT.font);

	button.Texture = DF:NewImage(button, [[Interface\Scenarios\ScenarioIcon-Combat]], 10, 10, nil, nil, nil)
	button.Texture:SetTexCoord (0.23, 0.87, 0.15, 0.73)
	button.Texture:SetPoint("RIGHT", button, "LEFT", -2, -1)
	button.Texture:SetSize (10, 10)
	button:SetText("Name Error")
	
	--MF.BossButtons[i].bossIndex = i + 1
	
	--MF.BossButtons[i]:Hide()
	
	PLT.MF.PlayerButtons[i] = button
end

local function getPlayerButtonText(i)
	local playerName = rollStorage.sortedRolls[i].player
	local rollType = rollStorage.sortedRolls[i].rollType
	local playerRolls
	if rollType == "MS" then
		playerRolls = rollStorage.playerRollDB.MS[playerName]
	else
		playerRolls = rollStorage.playerRollDB.OS[playerName]
	end
	
	local rollText = ""
	for i, roll in ipairs(playerRolls) do
		rollText = rollText .. " " .. roll
	end
	
	local text = PLT:NS(playerName) .. rollText
	
	PLT.MF.PlayerButtons[i]:SetText(text)
end

function PLT:UpdatePlayerButton(i)
	PLT.MF.PlayerButtons[i]:Show()
	--PLT.MF.PlayerButtons[i]:SetText(rollStorage.sortedRolls[i])
	getPlayerButtonText(i)
	PLT.MF.PlayerButtons[i].Texture:SetVertexColor(1, 1, 1, .4)
end

local function createRollDB(rollDB, roll, name)
	--I'm sure there is a better way to do this
	--Don't really care
	if not rollDB.rolls then rollDB.rolls = {} end
	if not rollDB.rolls[roll] then rollDB.rolls[roll] = {} end
	if not rollDB.rolls[roll].players then rollDB.rolls[roll].players = {} end
	if not rollDB.rolls[roll].players[name] then rollDB.rolls[roll].players[name] = {} end
	rollDB.rolls[roll].players[name] = true
	return rollDB.rolls[roll]
end

local function insertNameRoll(t, name, num)
	if not t[name] then
		t[name] = {}
		if not t.rollCount then t.rollCount = 0 end
		t.rollCount = t.rollCount + 1
	end
	table.insert(t[name], num)
	return t[name]
end

function PLT:UpdatePickWinnerButtonText(tie)
	if tie then
		PLT.MF.PickWinner:SetText(L["S_REROLL"])
		PLT:ButtonState(PLT.MF.PickWinner, true)
	else
		if rollStorage.winner then
			PLT.MF.PickWinner:SetText(L["S_NEW_WINNER"])
		else
			PLT.MF.PickWinner:SetText(L["S_PICK_WINNER"])
		end
		
		if rollStorage.selectedPlayer == rollStorage.winner then
			PLT:ButtonState(PLT.MF.PickWinner, false)
		else
			PLT:ButtonState(PLT.MF.PickWinner, true)
		end
	end
	
	if PLT.MF.hideTimer then PLT:ButtonState(PLT.MF.PickWinner, false) end
	
end

local function shouldAllowRoll(rollType, rollName)
	if not rollStorage.playerRollDB[rollType][rollName] then return true end
	
	if rollStorage.playerRollDB.ties[rollType][rollName] then
		local allowRoll
		local rollDB = rollStorage.playerRollDB[rollType]
		local curPlayerRolls = #rollDB[rollName]
		for tiedPlayer, _ in pairs(rollStorage.playerRollDB.ties[rollType][rollName]) do
			if #rollDB[tiedPlayer] >= curPlayerRolls then
				return true
			end
		
		end
	end
end

function EF:CHAT_MSG_SYSTEM(...)
	arg1 = ...;

	local rollName,rollNum,minRoll,maxRoll = arg1:match(RollPattern)
	
	if not (rollName and rollNum and minRoll and maxRoll) then return end
	
	rollNum = tonumber(rollNum)
	minRoll = tonumber(minRoll)
	maxRoll = tonumber(maxRoll)
	
	PLT:DebugPrint(rollName .. " has rolled " .. rollNum)
	
	if not minRoll == 1 then return end
	
	local rollType
	if maxRoll == MSNum then
		rollType = 'MS'
	elseif maxRoll == OSNum then
		rollType = 'OS'
	else return end
	
	--Return if they've already rolled for this roll type AND there is no tie. Continue if there is a tie.
	
	if not shouldAllowRoll(rollType, rollName) then return end

	if not rollStorage.playerRealm[rollName] then
		local _, playerRealm = UnitName(rollName)
		if playerRealm then rollStorage.playerRealm[rollName] = playerRealm end
	end
	
	local rollDB
	local nameRolls
	
	--Add the roll number to the player's roll list
	if maxRoll == MSNum then
		rollDB = rollStorage.MSRolls
		nameRolls = insertNameRoll(rollStorage.playerRollDB.MS, rollName, rollNum)
	else
		rollDB = rollStorage.OSRolls
		nameRolls = insertNameRoll(rollStorage.playerRollDB.OS, rollName, rollNum)
	end
	
	for i, roll in pairs(nameRolls) do
		rollDB = createRollDB(rollDB, roll, rollName)
	end
	
	-- if debugging then
		-- if not PLTraderDB.debug then PLTraderDB.debug = {} end
		-- PLTraderDB.debug = rollStorage
	-- end
	
	PLT.MF.RC.text:SetText("")
	PLT:ButtonState(PLT.MF.PickWinner, true)
	PLT:CreateSortedNameList()
	PLT:UpdatePickWinnerButtonText(getTieDB()) --Updates the win button to the ask for reroll text
	
end


function EF:CHAT_MSG_LOOT(...)

    if not IsInInstance() then return end

	local message, _, _, _, looter = ...;
	
	--Exit if you aren't the looter
	if not (PLT:NS(looter) == PLT:NS(UnitName('player'))) then return end
	
	local lootedItem = message:match(SelfLootPattern)
	if not lootedItem then return end
	PLT:DebugPrint("You Looted " .. lootedItem)

	PLT:CheckItem(lootedItem)
end



function PLT:UpdateNameWinnerSpec(tie)
	local typeText = L["S_CURRENTLY_SELECTED"]
	local name = rollStorage.winner or rollStorage.selectedPlayer
	local spec = rollStorage.playerSpec[name]
	
	if rollStorage.winner then
		typeText = L["S_WINNER_SELECTED"]
	elseif tie then
		name = L["S_TIE_SELECTED"]
	end
	
	--PLT:DebugPrint(spec)
	
	if not tie or rollStorage.winner then name = PLT:ColorName(name) end
	
	if not tie and spec then name = name.." - "..spec end

	PLT.MF.RC.nameText:SetText(typeText .. name)

end

function EF:INSPECT_READY(...)
	local guid = select(1, ...)
	local _, _, _, _, _, name, realm = GetPlayerInfoByGUID(guid)
	local spec = GetInspectSpecialization(name)
	_, spec = GetSpecializationInfoByID(spec)
	rollStorage.playerSpec[PLT:NS(name)] = spec
	if PLT:NS(name) == rollStorage.selectedPlayer then
		PLT:UpdateNameWinnerSpec(getTieDB())
		EF:UnregisterEvent("INSPECT_READY")
	end
end

function EF:TRADE_SHOW()
	--IgnoreSlots = {}
	--print("TRADE SHOW")
	EF:RegisterEvent("TRADE_CLOSED")
	
	local tradeReceiver = TradeFrameRecipientNameText:GetText()
	-- print(TradeReceiver)
	-- print(Winner)
	if string.match(tradeReceiver, rollStorage.winner) then
		C_Timer.After(0.1, function() PLT:TradeOpened() end)
	else
		PLT:Print(L["S_WRONG_PERSON_WARNING"])
	end
end

function EF:TRADE_CLOSED()
	EF:UnregisterEvent("TRADE_CLOSED")
	C_Timer.After(0.5, function() PLT:TradeClosed() end)
end


function PLT:CheckIfTradeable(item, manual)

    dbpr("Checking", item)
	local canTrade, _, _, isBoE = PLT:IsTradeable(item)
	
    --canTrade = true

	-- PLT:DebugPrint(canTrade)
	-- PLT:DebugPrint(item)
    --dbpr("Checking", item)
	
	if manual and not canTrade and not isDebugging() then PLT:Print(L["S_NOT_TRADABLE"]) end
	
    --don't auto show BoEs
    if isBoE and not manual then return end

    --PLTrader.db.profile.rclc.stopplt
	if canTrade or (manual and isDebugging()) then

        --Don't open up if RCLootCouncil dialog should open
        if not manual and PLTrader.db.profile.rclc.stopplt and (RCLootCouncil and RCLootCouncil.mldb.rejectTrade and IsInRaid()) then
            dbpr("... but it shouldn't open because RCLootCouncil dialog is enabled. Saving previous link.")
			itemStorage.prevItemLink = item
            return
        end

		if (PLTrader.db.profile.trader.autoshow or manual) and not PLTIsOpen() then
			dbpr("And the addon should open")
            itemStorage.curItemLink = item
			itemStorage.prevItemLink = itemStorage.curItemLink
			PLT:OpenMainFrame()
		else
            dbpr("... but it shouldn't open. Saving previous link.")
			itemStorage.prevItemLink = item
		end
	else
        dbpr("You can't trade it")
    end
end

function PLT:CheckItem(itemLink, manual)
	PLT:GetChannel()
	--Exit if not in a group and not debugging
	if not PLTrader.channel and not PLTrader.db.profile.general.debugging then return end 
	
	local item = Item:CreateFromItemLink(itemLink)
	
	item:ContinueOnItemLoad(function() --did this to try and debug window not showing up?
		itemLink = item:GetItemLink()
		local _, itemLink, itemRarity, _, _, type, subType, _, itemEquipLoc, _, _, ID, subID, bindType, _, _, _ = GetItemInfo(itemLink)
		--Has to be Epic, be equippable, and BoP to show. Return if not.

        --PLT:DebugPrint(itemLink, "EquipLoc:", itemEquipLoc, "bindType", bindType, "type", type, ID, "suType", subType, subID)

		if not manual and (bindType ~= 1 or itemRarity < 4) then PLT:DebugPrint("It should not be rolled on.") return end

        PLT:DebugPrint("It's BoP and high enough quality.")
		
		itemStorage.checkItem = itemLink
		
		--Delay to make sure it's in your bag, Not the best way?
        if manual then
            PLT:CheckIfTradeable(itemLink, true)
        else
            C_Timer.After(2, function() PLT:CheckIfTradeable(itemLink) end)
        end

	end)
end

function PLT:GetChannel()
	PLTrader.channel = nil
	if IsInGroup() then
		if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
			PLTrader.channel = 'INSTANCE_CHAT'
			return true
		elseif IsInRaid() then
			PLTrader.channel = 'RAID'
			return true
		else
			PLTrader.channel = 'PARTY'
			return true
		end
	end
end

function PLT:OpenDragFrame()
	if not PLT.MF then
		PLT:CreateMainFrames()
	end
	
	PLT.MF:Show()
	PLT.MF.SC:Hide()
	PLT.MF.RC:Hide()
	PLT.MF.DC:Show()
end

function PLT:OpenMainFrame()
	if not PLT.MF then
		PLT:CreateMainFrames()
	end
	if debugging and not PLT.MF.RC.DC then
		PLT:CreateDebugFrames()
	end
	
	PLT:UpdateLootIcon(itemStorage.curItemLink)
	PLT.MF:Show()
	PLT.MF.SC:Show()
	PLT.MF.RC:Hide()
	PLT.MF.DC:Hide()
	PLT:UpdateExampleText()
end

function PLT:TradeClosed()
	local canTrade, bag, slot = PLT:IsTradeable(itemStorage.curItemLink)
	if canTrade then
		PLT:Print(L["S_TRADE_FAILED_CHECK"])
		EF:RegisterEvent("TRADE_CLOSED")
	else
		PLT:HideMainFrame()
		PLT:Print(L["S_TRADE_COMPLETE"])

	end
end

function PLT:TradeOpened()
	if PLT.MF and PLT.MF.fromRCLC then return end
    local canTrade, bag, slot = PLT:IsTradeable(itemStorage.curItemLink) --Get Position of Aquired Item. Start at beginning.

	if canTrade then
		UseContainerItem(bag,slot)
	else
		local canTrade, bag, slot = PLT:IsTradeable(itemStorage.backupItemLink) --item isn't found the from the other item link sometimes? Maybe this fixes? Haven't tested
		if canTrade then
			UseContainerItem(bag,slot)
		else
			PLT:Print(L["S_ITEM_NOT_FOUND"])
		end
	end

end

local function updateCheckText(checkButton, text)
	_G[checkButton:GetName() .. 'Text']:SetText(text);
	local stringLen = string.len(text)
	local clickOffset = stringLen * 8 * -1
	checkButton:SetHitRectInsets(0,clickOffset,0,0)
end

local checkBoxCount = 1;
local function createCheckbutton(parent, anchorFrame, anchor1, anchor2, xoff, yoff, text, setting, tooltipText)
	checkBoxCount = checkBoxCount + 1;
	
	local checkButton = CreateFrame("CheckButton", "LPTCheckButton" .. checkBoxCount, parent, "ChatConfigCheckButtonTemplate");
	checkButton:SetPoint(anchor1, anchorFrame, anchor2, xoff, yoff);
	_G[checkButton:GetName() .. 'Text']:SetText(text);
	_G[checkButton:GetName() .. 'Text']:SetFont(PLT_FONT, 12)

	local stringLen = string.len(text)
	local clickOffset = stringLen * 8 * -1 --Not really sure how to get the actual string of a check box's length. Seems acurate enough.
	
	checkButton:SetHitRectInsets(0,clickOffset,0,0)
	--_G[checkButton:GetName() .. 'Text']:SetFont(PLT.font);
	checkButton:SetChecked(PLTrader.db.profile.trader[setting])
	checkButton:SetScript("OnClick", 
	  function()
		PLTrader.db.profile.trader[setting] = not PLTrader.db.profile.trader[setting]
		checkButton:SetChecked(PLTrader.db.profile.trader[setting])
		PLT:UpdateExampleText()
	  end
	);
	
	if tooltipText then
		checkButton:SetScript("OnEnter", function() PLT:ToggleInfoTooltip(true, tooltipText) end)
		checkButton:SetScript("OnLeave", function() PLT:ToggleInfoTooltip(false) end)
		
	end
	
	--checkButton:SetBackdrop ({edgeFile = [[Interface\Buttons\WHITE8X8]], edgeSize = 2, bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tileSize = 64, tile = true})
	
	--[[
	checkButton.texture = checkButton:CreateTexture()
	checkButton.texture:SetTexture("Interface\\AddOns\\PersonalLootTrader\\libs\\DF\\background")

	checkButton.texture:SetAllPoints(checkButton)
	checkButton.texture:SetWidth(5)
	checkButton.texture:SetHeight(5)
	
	checkButton:SetNormalTexture(checkButton.texture)
	]]--
	
	return checkButton;
end

function PLT:CreateTextButton(name, text, parent, anchorFrame, anchor1, anchor2, xoff, yoff, width, height)
	local button = CreateFrame ("button", "PLT"..name, parent, "BackdropTemplate")
	button:SetPoint(anchor1, anchorFrame, anchor2, xoff, yoff)
	button:SetWidth(width)
	button:SetHeight(height)
	button:SetText(text)
	button:SetBackdrop ({edgeFile = [[Interface\Buttons\WHITE8X8]], edgeSize = 2, bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tileSize = 64, tile = true})

	button:SetNormalFontObject(PLT.font);
	
	button:SetBackdropColor (.4, .4, .4, 1)
	button:SetBackdropBorderColor (.7, .7, .7, 1)
	
	return button
end

function PLT:NS(name)
	name = gsub(name, "%-[^|]+", "")
	return name
end

function PLT:UpdateFramePositions()
	local scale = self.db.profile.general.scale
	if PLT.MF then
		PLT.MF:SetScale(scale)
		PLT.MF:ClearAllPoints()
		--PLT.MF:SetPoint ("BOTTOMLEFT", tonumber(self.db.profile.trader.xoff/scale), tonumber(self.db.profile.trader.yoff/scale))
		PLT.MF:SetPoint ("TOPLEFT", PLT.MF:GetParent(), "BOTTOMLEFT", tonumber(self.db.profile.trader.xoff/scale), tonumber(self.db.profile.trader.yoff/scale))
	end
	
	if PLT.RF then
		PLT.RF:SetScale(scale)
		PLT.RF:ClearAllPoints()
		PLT.RF:SetPoint ("TOPLEFT", PLT.RF:GetParent(), "BOTTOMLEFT", tonumber(self.db.profile.roller.xoff/scale), tonumber(self.db.profile.roller.yoff/scale))
	end

end

function PLT:CreateBackgroundFrame(width, height)
	
	local f = CreateFrame ("frame", nil, UIParent, "BackdropTemplate")
	f:SetSize(width, height)
	f:SetBackdrop ({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16, edgeFile = [[Interface\ButtoPLT:NS\WHITE8X8]], edgeSize = 2})
	f:SetBackdropColor (0, 0, 0, .8)
	f:SetBackdropBorderColor (0, 0, 0, 1)
	return f

end

function PLT:CreateMainFrames()
	--main frame
	PLT.MF = PLT:CreateBackgroundFrame(MAIN_FRAME_WIDTH, MAIN_FRAME_HEIGHT)
	
	--PLT:SetDefaultLocation()
	PLT:UpdateFramePositions()
	
	PLT.MF:EnableMouse (true)
	PLT.MF:SetMovable (true)
	
	PLT.MF:Hide()
	PLT.MF:SetScript("OnMouseDown", function(self, button)
	  if button == "LeftButton" and not self.isMoving then
	   self:StartMoving();
	   self.isMoving = true;
	  end
	end)
	PLT.MF:SetScript("OnMouseUp", function(self, button)
	  if button == "LeftButton" and self.isMoving then
	   self:StopMovingOrSizing();
	   self.isMoving = false;
	   
	   --local point, relativeTo, relativePoint, xOfs, yOfs = PLT.MF:GetPoint(1)
	   local scale = PLTrader.db.profile.general.scale
	   
	   PLTrader.db.profile.trader.xoff = PLT:Round(self:GetLeft()*scale, 0)
	   PLTrader.db.profile.trader.yoff = PLT:Round(self:GetTop()*scale, 0)
	   
	  end
	end)
	PLT.MF:SetScript("OnHide", function(self)
	  if ( self.isMoving ) then
	   self:StopMovingOrSizing();
	   self.isMoving = false;
	  end
	end)

	PLT.MF.PlayerButtons = {}

	--title bar
	PLT.MF.TitleBar = CreateFrame ("frame", "PLTTitleBar", PLT.MF, "BackdropTemplate")
	PLT.MF.TitleBar:SetPoint ("topleft", PLT.MF, "topleft", 2, -3)
	PLT.MF.TitleBar:SetPoint ("topright", PLT.MF, "topright", -2, -3)
	PLT.MF.TitleBar:SetHeight (20)
	PLT.MF.TitleBar:EnableMouse (false)
	PLT.MF.TitleBar:SetBackdrop ({edgeFile = [[Interface\Buttons\WHITE8X8]], edgeSize = 2, bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tileSize = 64, tile = true})
	PLT.MF.TitleBar:SetBackdropColor (.2, .2, .2, 1)
	PLT.MF.TitleBar:SetBackdropBorderColor (0, 0, 0, .5)

	-- title text
	PLT.MF.Title = PLT.MF.TitleBar:CreateFontString ("PLTTitleText", "overlay", "GameFontNormal")
	PLT.MF.Title:SetPoint ("CENTER", PLT.MF.TitleBar, "CENTER")
	PLT.MF.Title:SetTextColor (.8, .8, .8, 1)
	PLT.MF.Title:SetText (L["S_ADDON_NAME"])

	PLT.MF.AnchorFrame = CreateFrame ("frame", nil, PLT.MF)
	PLT.MF.AnchorFrame:SetAllPoints()
	PLT.MF.AnchorFrame:SetFrameLevel (PLT.MF:GetFrameLevel()+3)

	--Start Button Container
	
	PLT.MF.SC = CreateFrame("Frame", "#SC", PLT.MF)
	PLT.MF.SC:Hide()

	local borderSize = 2
	--loot border
	PLT.MF.LootBorder = CreateFrame ("frame", nil, PLT.MF.SC, "BackdropTemplate")
	PLT.MF.LootBorder:SetPoint ("TOPLEFT", PLT.MF.TitleBar, "BOTTOMLEFT", 2, -5)
	PLT.MF.LootBorder:SetHeight(70)
	PLT.MF.LootBorder:SetWidth(70)
	PLT.MF.LootBorder:EnableMouse (false)
	PLT.MF.LootBorder:SetBackdrop ({edgeFile = [[Interface\Buttons\WHITE8X8]], edgeSize = borderSize*2, bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tileSize = 64, tile = true})
	PLT.MF.LootBorder:SetBackdropColor (1, 1, 1, 1)
	PLT.MF.LootBorder:SetBackdropBorderColor (1, 1, 1, .5)

	--loot button
	PLT.MF.Loot = CreateFrame ("button", nil, PLT.MF.LootBorder)
	PLT.MF.Loot:SetPoint("TOPLEFT", PLT.MF.LootBorder, "TOPLEFT", borderSize+1, -1*borderSize-1)
	PLT.MF.Loot:SetWidth(64)
	PLT.MF.Loot:SetHeight(64)

	PLT.MF.Loot:SetNormalFontObject("GameFontNormal")
	PLT.MF.Loot:SetScript("OnEnter", function() PLT:ShowItemTooltip() end)
	PLT.MF.Loot:SetScript("OnLeave", function() PLT:HideItemTooltip() end)

	--close button
	PLT.MF.Close = CreateFrame ("button", nil, PLT.MF)
	PLT.MF.Close:SetPoint ("right", PLT.MF.TitleBar, "right", -2, 0)
	PLT.MF.Close:SetSize (16, 16)
	PLT.MF.Close:SetNormalTexture (PLT.ICONPATH)
	PLT.MF.Close:SetHighlightTexture (PLT.ICONPATH)
	PLT.MF.Close:SetPushedTexture (PLT.ICONPATH)
	PLT.MF.Close:GetNormalTexture():SetTexCoord (0, 16/128, 0, 1)
	PLT.MF.Close:GetHighlightTexture():SetTexCoord (0, 16/128, 0, 1)
	PLT.MF.Close:GetPushedTexture():SetTexCoord (0, 16/128, 0, 1)
	PLT.MF.Close:SetAlpha (0.7)
	PLT.MF.Close:SetScript ("OnClick", function() PLT:HideMainFrame() end)

	--options button
	PLT.MF.Options = CreateFrame ("button", nil, PLT.MF.SC)
	PLT.MF.Options:SetPoint ("right", PLT.MF.Close, "left", -2, 0)
	PLT.MF.Options:SetSize (16, 16)
	PLT.MF.Options:SetNormalTexture (PLT.ICONPATH)
	PLT.MF.Options:SetHighlightTexture (PLT.ICONPATH)
	PLT.MF.Options:SetPushedTexture (PLT.ICONPATH)
	PLT.MF.Options:GetNormalTexture():SetTexCoord (48/128, 64/128, 0, 1)
	PLT.MF.Options:GetHighlightTexture():SetTexCoord (48/128, 64/128, 0, 1)
	PLT.MF.Options:GetPushedTexture():SetTexCoord (48/128, 64/128, 0, 1)
	PLT.MF.Options:SetAlpha (0.7)
	PLT.MF.Options:SetScript ("OnClick", function()
		InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
		InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
	end)
	
	--Fonts
	
	PLT.MF.playerFont = CreateFont("PLTPlayerFont")
	PLT.MF.playerFont:CopyFontObject("GameFontNormal");
	PLT.MF.playerFont:SetTextColor(1, 1, 1, 1);
	PLT.MF.playerFont:SetFont(PLT_FONT, 12, "")
	PLT.MF.playerFont:SetJustifyH("LEFT")
	
	PLT.MF.title = CreateFont("PLTTitleFont")
	PLT.MF.title:CopyFontObject("GameFontNormal");
	PLT.MF.title:SetTextColor(1, 1, 1, 1.0);
	PLT.MF.title:SetFont(PLT_FONT, 12, "")
	PLT.MF.title:SetJustifyH("CENTER")
	
	--Group Roll
	
	PLT.MF.StartRoll = PLT:CreateTextButton("Roll", L["S_START_ROLL"], PLT.MF.SC, PLT.MF.LootBorder, "TOPLEFT", "TOPRIGHT", 5, 0, 75, 28)
	PLT.MF.StartRoll:SetScript ("OnClick", function() PLT:StartGroupRoll() end)
	-- PLT.MF.StartRoll:SetScript("OnEnter", function() PLT:ToggleInfoTooltip(true, L["S_GROUP_ROLL_DESC"]) end)
	-- PLT.MF.StartRoll:SetScript("OnLeave", function() PLT:ToggleInfoTooltip(false) end)
	

	--Give to button
	PLT.MF.GiveTo = PLT:CreateTextButton("Give", L["S_GIVE"], PLT.MF.SC,PLT.MF.StartRoll, "TOPLEFT", "BOTTOMLEFT", 0, -5, 75, 36)
	PLT.MF.GiveTo:SetScript ("OnClick", function() PLT:CustomWinner() end)
	PLT.MF.GiveTo:SetScript("OnEnter", function() PLT:ToggleInfoTooltip(true, L["S_GIVE_DESC"]) end)
	PLT.MF.GiveTo:SetScript("OnLeave", function() PLT:ToggleInfoTooltip(false) end)

	
	--Keep Button
	PLT.MF.Keep = PLT:CreateTextButton("Keep", L["S_KEEP"], PLT.MF.SC, PLT.MF.StartRoll, "TOPLEFT", "TOPRIGHT", 5, 0, 75, 70)
	PLT.MF.Keep:SetScript ("OnClick", function() PLT:HideMainFrame() end)
	
	local checkSpacing = -5
	
	--Check Boxes
	PLT.MF.Auto = createCheckbutton(PLT.MF.SC, PLT.MF.LootBorder, "TOPLEFT", "BOTTOMLEFT", 0, checkSpacing, L["S_AUTO"], "autoAssign", L["S_GROUP_ROLL_DESC"])
	PLT.MF.allowOS = createCheckbutton(PLT.MF.SC, _G[PLT.MF.Auto:GetName() .. 'Text'] , "LEFT", "RIGHT", 5, -2, PLT.db.profile.trader.osabbrev, "allowOS", L["S_OS_DESC"])
	PLT.MF.Timer = createCheckbutton(PLT.MF.SC, _G[PLT.MF.allowOS:GetName() .. 'Text'] , "LEFT", "RIGHT", 5, -2, L["S_TIMER"], "usetimer", L["S_USE_TIMER_DESC"])
	
	PLT.MF.SC.text = PLT.MF.SC:CreateFontString ("$parentSCText", "overlay", "PLTButtonFont")
	
	PLT.MF.SC.text:SetWidth(MAIN_FRAME_WIDTH-10)
	PLT.MF.SC.text:SetHeight(100)
	PLT.MF.SC.text:SetJustifyH("LEFT")
	--PLT.MF.SC.text:SetNonSpaceWrap(false)
	PLT.MF.SC.text:SetJustifyV("TOP")
	PLT.MF.SC.text:SetPoint ("TOPLEFT", PLT.MF.Auto, "BOTTOMLEFT", 2, -5)
	
	PLT.MF.SC.text:SetTextColor(1, 1, 1, 1)
	
	--Contains frames that show when a roll starts. RC = roll container
	PLT.MF.RC = CreateFrame("Frame", nil, PLT.MF.TitleBar)
	PLT.MF.RC:Hide()
	
	--Name Text
	PLT.MF.RC.text = PLT.MF.RC:CreateFontString ("$parentTitle", "overlay", "GameFontNormal")
	PLT.MF.RC.text:SetWidth(MAIN_FRAME_WIDTH)
	PLT.MF.RC.text:SetHeight(50)
	PLT.MF.RC.text:SetJustifyH("CENTER")
	PLT.MF.RC.text:SetJustifyV("CENTER")
	PLT.MF.RC.text:SetPoint ("TOP", PLT.MF.Title, "BOTTOM", 0, -35)
	PLT.MF.SC.text:SetFont(PLT_FONT, 12)
	PLT.MF.RC.text:SetTextColor (.8, .8, .8, 1)

	--Pick Winner Button
	PLT.MF.PickWinner = PLT:CreateTextButton("Pick Winner", L["S_PICK_WINNER"], PLT.MF.RC, PLT.MF, "BOTTOMLEFT", "BOTTOMLEFT", 5, 5, 100, 22)
	PLT.MF.PickWinner:SetScript ("OnClick", function() PLT:PickWinnerPress() end)
	
	--Text that shows the name of the winner
	PLT.MF.RC.nameText = PLT.MF.RC:CreateFontString (nil, "overlay", "GameFontNormal")
	PLT.MF.RC.nameText:SetWidth(MAIN_FRAME_WIDTH-10)
	PLT.MF.RC.nameText:SetHeight(25)
	PLT.MF.RC.nameText:SetJustifyH("LEFT")
	PLT.MF.RC.nameText:SetJustifyV("TOP")
	PLT.MF.RC.nameText:SetPoint ("BOTTOMLEFT", PLT.MF.PickWinner, "TOPLEFT", 0, 10)
	PLT.MF.RC.nameText:SetFont(PLT_FONT, 12)
	PLT.MF.RC.nameText:SetTextColor (.8, .8, .8, 1)
	
	--Trade/follow button
	PLT.MF.Trade = PLT:CreateTextButton("Trade", L["S_INITIATE_TRADE"], PLT.MF.RC, PLT.MF.PickWinner, "LEFT", "RIGHT", 5, 0, 100, 22)
	PLT.MF.Trade:SetScript ("OnClick", function() PLT:InitiateTrade() end)
	
	
	--Anchor Frame
	PLT.MF.AnchorFrame = CreateFrame ("frame", nil, PLT.MF)
	PLT.MF.AnchorFrame:SetAllPoints()
	PLT.MF.AnchorFrame:SetFrameLevel (PLT.MF:GetFrameLevel()+3)

	--Progress Bar
	PLT.MF.ProgressBar = DF:CreateBar (PLT.MF.AnchorFrame, nil, 125, 22, 50)
	--PLT.MF.ProgressBar:SetPoint (80, -83)
	PLT.MF.ProgressBar:SetPoint ("TOPLEFT", PLT.MF.PickWinner, "TOPRIGHT", 5, 0)
	PLT.MF.ProgressBar.fontsize = 11
	PLT.MF.ProgressBar.fontface = "GameFontNormal"
	PLT.MF.ProgressBar.fontcolor = "white"
	PLT.MF.ProgressBar.color = "gray"	
	PLT.MF.ProgressBar:EnableMouse (false)
	PLT.MF.ProgressBar:Hide()

	PLT.MF.ProgressBar.timer_texture:SetTexture ([[Interface\Buttons\WHITE8X8]])
	
	--Drag Container
	PLT.MF.DC = CreateFrame("Frame", nil, PLT.MF.TitleBar)
	PLT.MF.DC:Hide()
	
	PLT.MF.DC.LootBorder = CreateFrame ("frame", nil, PLT.MF.DC, "BackdropTemplate")
	PLT.MF.DC.LootBorder:SetPoint ("CENTER", PLT.MF, "CENTER", 0, 0)
	PLT.MF.DC.LootBorder:SetHeight(70)
	PLT.MF.DC.LootBorder:SetWidth(70)
	PLT.MF.DC.LootBorder:EnableMouse (false)
	PLT.MF.DC.LootBorder:SetBackdrop ({edgeFile = [[Interface\Buttons\WHITE8X8]], edgeSize = borderSize*2, bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tileSize = 64, tile = true})
	PLT.MF.DC.LootBorder:SetBackdropColor (0, 0, 0, .8)
	PLT.MF.DC.LootBorder:SetBackdropBorderColor (1, 1, 1, .5)
	PLT.MF.DC.LootBorder:SetScript("OnEnter", function() PLT:DCUpdate(true) end)
	PLT.MF.DC.LootBorder:SetScript("OnLeave", function() PLT:DCUpdate() end)
	
	PLT.MF.DC.Text = PLT.MF.DC:CreateFontString ("PLTDCText", "overlay", "GameFontNormal")
	PLT.MF.DC.Text:SetFont(PLT_FONT, 8)
	PLT.MF.DC.Text:SetPoint("CENTER", PLT.MF.DC.LootBorder, "CENTER")
	PLT.MF.DC.Text:SetJustifyH("CENTER")
	PLT.MF.DC.Text:SetText(L["DRAG_EXPLAIN"])
	PLT.MF.DC.Text:SetTextColor (1, 1, 1, 1)
end

local function setTooltipWidth(isRoller)
	local scale = PLTrader.db.profile.general.scale
	PLT.infoTooltip.text:SetWidth((MAIN_FRAME_WIDTH*scale)-10)
	if isRoller then
		PLT.infoTooltip:SetWidth((PLT.infoTooltip.text:GetStringWidth()+10))
	else
		PLT.infoTooltip:SetWidth(MAIN_FRAME_WIDTH*scale)
		PLT.infoTooltip.text:SetWidth((MAIN_FRAME_WIDTH*scale)-10)
	end
end

local function createInfoTooltip()
	local scale = PLTrader.db.profile.general.scale
	local f = PLT:CreateBackgroundFrame(MAIN_FRAME_WIDTH*scale, MAIN_FRAME_HEIGHT*scale)
	f.text = f:CreateFontString(nil, "overlay", "GameFontNormal")
	f.text:SetFont(PLT_FONT, 10)
	f.text:SetPoint("TOPLEFT", f, "TOPLEFT", 5, -5)
	-- f.text:SetWidth((MAIN_FRAME_WIDTH*scale)-10)
	f.text:SetTextColor (1, 1, 1, 1)
	f.text:SetJustifyH("LEFT")
	return f
end

function PLT:ToggleInfoTooltip(show, text, rollerFrame)
	if show and (PLTrader.db.profile.general.tooltips or rollerFrame) then
		PLT.infoTooltip = PLT.infoTooltip or createInfoTooltip()
        PLT.infoTooltip.text:SetText(text)
		
		--local textHeight = PLT.infoTooltip.text:GetStringHeight();
		--PLT.infoTooltip:SetHeight(textHeight + 10)
		
		setTooltipWidth(rollerFrame)

		PLT.infoTooltip:Show()

        local textHeight = PLT.infoTooltip.text:GetStringHeight();
		PLT.infoTooltip:SetHeight(textHeight + 10)
		
		if rollerFrame then
			PLT.infoTooltip:SetPoint("TOPLEFT", rollerFrame, "TOPRIGHT", 10, 0)
		else
			PLT.infoTooltip:SetPoint("TOPLEFT", PLT.MF, "BOTTOMLEFT", 0, -5)
		end
	
		PLT.infoTooltip:Show()
	
	else
		if PLT.infoTooltip then PLT.infoTooltip:Hide() end
	end	
end

function PLT:UpdateTimerCheck(arg1)
	if PLT.MF and PLT.MF.Timer then
		PLT.MF.Timer:SetChecked(arg1)
	end
end

function PLT:ShowItemTooltip()
	--for i=1,tooltip:NumLines()do local mytext=_G["GameTooltipTextLeft"..i] local text=mytext:GetText()print(text)end
	--print("enter")
	GameTooltip:SetOwner (PLT.MF.Loot, "ANCHOR_LEFT")
	GameTooltip:SetHyperlink(itemStorage.curItemLink)
	--GameTooltip:SetText ("why", nil, nil, nil, nil, true)
	GameTooltip:Show()
	
	GameTooltip_ShowCompareItem(GameTooltip)
end

function PLT:HideItemTooltip()
	GameTooltip:Hide()
end

function PLT:UpdateLootIcon(ItemLink)
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, _, _, _,
	itemEquipLoc, iconFileDataID, _, itemClassID, itemSubClassID, bindType, expacID, _, 
	_ = GetItemInfo(ItemLink)

	PLT.MF.itemRarity = itemRarity
	itemStorage.backupItemLink = itemStorage.curItemLink
	itemStorage.curItemLink = itemLink
	
	PLT.itemToolTip = PLT.itemToolTip or CreateFrame("GameTooltip", "PLTScanningTooltip", nil, "GameTooltipTemplate")--PLT:CreateEmptyTooltip(50)
	PLT.itemToolTip:SetOwner(UIParent, "ANCHOR_NONE")
    PLT.itemToolTip:ClearLines()
	PLT.itemToolTip:SetHyperlink(ItemLink)
	
	itemStorage.realILVL = PLT:GetILVLFromTooltip(PLT.itemToolTip)
	if itemStorage.realILVL == nil then  
		itemStorage.realILVL = itemLevel
	end
	
	if itemRarity then
		r, g, b, hex = GetItemQualityColor(itemRarity)		
		PLT.MF.LootBorder:SetBackdropColor (r, g, b, 1)
		PLT.MF.LootBorder:SetBackdropBorderColor (r, g, b, .5)
	end
		
	local ntex = PLT.MF.Loot:CreateTexture()
	ntex:SetTexture(iconFileDataID)
	ntex:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	ntex:SetAllPoints()
	--ntex:SetVertexColor(1, 1, 1, 1)
	PLT.MF.Loot:SetNormalTexture(ntex)

end


function PLT:CreateDebugFrames()
	--Button Container
	
	PLT.MF.RC.DC = CreateFrame("Frame", nil, PLT.MF.RC)
	
	--MS Roll
	PLT.MF.AddMSRoll = PLT:CreateTextButton("MSDebug", "Add MS" , PLT.MF.RC.DC, PLT.MF, "TOPLEFT", "TOPRIGHT", 5, 0, 70, 20)
	PLT.MF.AddMSRoll:SetScript ("OnClick", function() PLT:DebugRollAdd(MSNum) end)
	PLT.MF.AddOSRoll = PLT:CreateTextButton("MSDebug", "Add OS" , PLT.MF.RC.DC, PLT.MF.AddMSRoll, "TOPLEFT", "BOTTOMLEFT", 0, -5, 70, 20)
	PLT.MF.AddOSRoll:SetScript ("OnClick", function() PLT:DebugRollAdd(MSNum-50) end)

end

local function someoneRolled()
	if rollStorage.playerRollDB.MS.rollCount or rollStorage.playerRollDB.OS.rollCount then
		return true
	end
end

function PLT:TimerDone()
	PLT.MF.hideTimer = nil
	PLT:ButtonState(PLT.MF.PickWinner, true)
	if someoneRolled() then
		PLT:PickWinnerPress()
	else
		PLT:Print(L["S_NO_ONE"])
		PLT:HideMainFrame()
	end
end

function PLT:EndGroupRoll()

	if rollStorage and rollStorage.inProgress and not rollStorage.winner then --If there is a winner then the message was already sent
		local addonMessage = "End~nil~"..MSNum
		C_ChatInfo.SendAddonMessage("PLTrader2", addonMessage, PLTrader.channel);
	end
	
	PLT:InitiateRollStorage()
	PLT.MF.RC:Hide()
	updateScrollFrame(PLT.MF.ScrollFrame)
end

function EF:PLAYER_LOGOUT()
	if not rollStorage then return end
	if not rollStorage.inProgress then return end

	local addonMessage = "End~nil~"..MSNum
	C_ChatInfo.SendAddonMessage("PLTrader2", addonMessage, PLTrader.channel);
	
	if not PLTrader.db.profile.trader.usetimer then
		if someoneRolled() and not rollStorage.winner then
			SendChatMessage(L["S_ADDON_ABREV"].. " " ..format (L["S_ROLL_CANCELLED"], itemStorage.curItemLink), PLTrader.channel)
		end
	else
		SendChatMessage(format (L["S_ROLL_CANCELLED"], itemStorage.curItemLink), PLTrader.channel)
	end
	
	
end

--hide the addon and reset it back to default
function PLT:HideMainFrame()
	
	if PLT.MF.hideTimer then
		PLT.MF.hideTimer:Cancel()
		PLT.MF.ProgressBar:Hide()
		PLT.MF.hideTimer = nil
	end
	
	PLT.MF.RC:Hide()	
	PLT.MF:Hide()
	
    PLT.MF.fromRCLC = false

	PLT:RegisterDefaultEvents()
	
	PLT:EndGroupRoll()
	
	PLT.MF.RC.text:SetText("")

	--IgnoreSlots = {}
end

local function getMSNumber(isExample)
	if not PLTrader.db.profile.trader.autoAssign then
		return 100
	end

	local num = START_NUM - 100
	repeat
		num = num + 100
	until not inUseNumbers[num]
	--if not isExample then PLT:AddUseNumber(num) end
	return num
end

function PLT:GetRollString(localMSNum, localOSNum, isExample)
	local message, rollStr
	local itemLink = "|cffa335ee[Item]|r"
	
	if not localMSNum then
		MSNum = getMSNumber(isExample)
		localMSNum = MSNum
		if PLTrader.db.profile.trader.allowOS then
			OSNum = MSNum - 50
			localOSNum = OSNum
		else
			OSNum = -1
		end
		
	end

	if localOSNum then
		message = PLTrader.db.profile.trader.MSOSMessage
		message = message:gsub("%[d1%]", PLTrader.db.profile.trader.msabbrev)
		message = message:gsub("%[d2%]", PLTrader.db.profile.trader.osabbrev)
		message = message:gsub("%[r1%]", localMSNum)
		message = message:gsub("%[r2%]", localOSNum)
	else
		message = PLTrader.db.profile.trader.MSMessage
		message = message:gsub("%[r1%]", localMSNum)
	end
	message = message:gsub(" 100", "") --Don't need to type 100 if the roll is 100 so remove it from the text
	--message = message:gsub('"/roll"', "/roll")

	if itemStorage.curItemLink and not isExample then
		itemLink = itemStorage.curItemLink
	end
	
	if PLTrader.db.profile.trader.usetimer then
		message = message.." "..format(L["S_SECOND_COUNT"], PLTrader.db.profile.trader.length)
	end
	
	rollStr = L["S_ADDON_ABREV"].. " ".. itemLink .." ".. message
	return rollStr
end

--this is so bad why did I make it like this
function PLT:UpdateExampleText(isRestart)
	if (PLT.MF and PLT.MF.SC.text:IsVisible()) or (isRestart and PLT.MF:IsVisible()) then
		local rollStr = PLT:GetRollString(nil, nil, true) --nil, nil, true = get normal roll numbers but use example item link to save space
		PLT.MF.SC.text:SetText(rollStr)
	end
	if PLT.MF then
		updateCheckText(PLT.MF.allowOS, PLT.db.profile.trader.osabbrev)
	end
end

function PLT:InitiateRollStorage()
	rollStorage = {playerRealm = {}, playerClass = {}, playerSpec = {}, MSRolls = {}, OSRolls = {}, sortedRolls = {}, ignoreNames = {}, playerRollDB = {ties = {MS = {}, OS = {}}, MS = {}, OS = {}}}
end

function PLT:ButtonState(button, enable)
	if enable then
		local text = button:GetText()
		text = text:gsub("|c%x%x%x%x%x%x%x%x", '')
		text = text:gsub("|r", '')
		button:SetText(text)
		button:SetBackdropColor (.4, .4, .4, 1)
		button:SetBackdropBorderColor (.7, .7, .7, 1)
	else --disable
		button:SetText("|cff808080".. button:GetText() .. "|r")
		button:SetBackdropColor (.2, .2, .2, 1)
		button:SetBackdropBorderColor (.4, .4, .4, 1)
	end
end


function PLT:CustomWinner(name)
	
    if not name then
        name = (UnitInParty('target') or UnitInRaid('target')) and select(1, UnitName("target")) or nil
    end

	if name then
	
		PLT.MF.manualSelect = false
		PLT.MF.SC:Hide()
		PLT.MF.RC:Show()
		PLT.MF.PickWinner:SetText(L["S_PICK_WINNER"])
		PLT:ButtonState(PLT.MF.PickWinner, false)
		
		PLT:InitiateRollStorage()

		local str = format(_G.RANDOM_ROLL_RESULT, name, 100, 1, 100)
		MSNum = 100
		PLT:GetRaidMemberClasses()
		EF:CHAT_MSG_SYSTEM(str)
		PLT:PickWinnerPress(true)
	else
		PLT:Print(L["S_NO_TARGET"] )
	end

end

function PLT:StartGroupRoll()

	if not PLT:GetChannel() and not isDebugging() then
		PLT:Print(L["S_ONLY_GROUP"])
		return
	end
	
	PLT.MF.manualSelect = false
	
	PLT.MF.SC:Hide()
	PLT.MF.RC:Show()
	PLT.MF.Trade:Hide()
	--PLT.MF.PickWinner:Hide()
	PLT.MF.RC.text:SetText(L["S_NO_ROLL"])
	
	PLT.MF.RC.nameText:SetText("")
	
	PLT:InitiateRollStorage()
	rollStorage.inProgress = true
	
	PLT:GetRaidMemberClasses()
	
	EF:RegisterEvent("CHAT_MSG_SYSTEM")
	
	local timerLength
	if PLTrader.db.profile.trader.usetimer then
		timerLength = PLTrader.db.profile.trader.length + 1
	else
		timerLength = -1
	end
	
	local addonMessage = "Start"..d..itemStorage.curItemLink..d..timerLength..d..MSNum..d..OSNum..d..PLTrader.db.profile.trader.msabbrev..d..PLTrader.db.profile.trader.osabbrev

	C_ChatInfo.SendAddonMessage("PLTrader2", addonMessage, PLTrader.channel);

	if timerLength > -1 then
		PLT.MF.ProgressBar:Show()
		PLT.MF.ProgressBar:SetTimer (timerLength)
		PLT.MF.hideTimer = C_Timer.NewTimer(PLTrader.db.profile.trader.length+1, function() PLT:TimerDone() end)
		-- PLT.MF.PickWinnerButton:Show()		
		-- PLT.MF.PickWinnerButton:SetText(L["S_CANCEL"])
	end
	
	PLT.MF.PickWinner:SetText(L["S_PICK_WINNER"])
	PLT:ButtonState(PLT.MF.PickWinner, false)
	
	--local message = PLT:GetRollString(nil, nil, false)
	--if not isDebugging() then SendChatMessage(message, PLTrader.channel) else print(message) end
	--SendChatMessage(message, PLTrader.channel)

end

function PLT:SendStartMessage()
	local message = PLT:GetRollString(nil, nil, false)
	--if not isDebugging() then SendChatMessage(message, PLTrader.channel) else print(message) end
	SendChatMessage(message, PLTrader.channel)
end


function PLT:SetDefaultLocation(force)
	if not PLTrader.db.profile.trader.xoff or force then
		local scale = PLTrader.db.profile.general.scale
		
		PLTrader.db.profile.trader.xoff = (GetScreenWidth()/2)-((MAIN_FRAME_WIDTH/2)*scale);
		PLTrader.db.profile.trader.yoff = (GetScreenHeight()/2)+((MAIN_FRAME_HEIGHT/2)*scale);
			
		PLTrader.db.profile.roller.xoff = (GetScreenWidth()/4)--*scale);
		PLTrader.db.profile.roller.yoff = (GetScreenHeight()-GetScreenHeight()/4)--*3)*scale;
	end
end

local function MyAddonCommands(msg, editbox)
	local capMsg = strupper(msg)
	if capMsg == 'OPTIONS' or capMsg == 'OPT' then
		InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
		InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
		
	elseif capMsg == "HELP" or capMSG == '' then
		 PLT:Print(L["S_SLASH_EXPLANATION"])
	elseif capMsg == "RESET" or capMsg == "RESETPOS" then
		
		PLT:Print(L["S_RESET_POS"])
		PLT:SetDefaultLocation(true)
		
		PLT:UpdateFramePositions()
	elseif capMsg == "TEST" and PLTrader.db.profile.general.debugging then
		PLT:RollTradeClosed(PLT:NS(UnitName('player')).." (*)")
	elseif not PLTIsOpen() then

		if not PLT:GetChannel() and not PLTrader.db.profile.general.debugging then
			PLT:Print(L["S_ONLY_GROUP"])
			return
		end
		
		if capMsg == "PREV" or capMsg == "PREVIOUS" or capMsg == "P" then
			if itemStorage.prevItemLink then
				PLT:CheckItem(itemStorage.prevItemLink, true) --true forces item to show up even though auto is not enabled
			else
				PLT:Print(L["S_NO_PREVIOUS"])
			end
		else
			local itemLink = msg:match("|%x+|Hitem:.-|h.-|h|r")
			if itemLink then
				PLT:CheckItem(itemLink, true)
			else
				PLT:Print(L["S_SLASH_EXPLANATION_TWO"])
			end
		end
	else
		PLT:Print(L["S_ONLY_ONE_INSTANCE"])
	end
end

SLASH_PLTraderCommand1 = '/plt'
SlashCmdList["PLTraderCommand"] = MyAddonCommands
